#include <iostream>
#include <CL/cl.hpp>
#include <omp.h>

#include <cmath>

#define PROBLEM_SIZE 100000000UL

int main(){
        //Search the compute platform first
        std::vector<cl::Platform> platforms;
        cl::Platform::get(&platforms);

        for(cl::Platform &p:platforms){
                std::cout << p.getInfo<CL_PLATFORM_NAME>() << std::endl;
        }
        //Choose the deafault platform
        cl::Platform myplat=platforms[0];
        
        //Search all compute devices in the platform
        std::vector<cl::Device> devices;
        myplat.getDevices(CL_DEVICE_TYPE_ALL, &devices);
        for(cl::Device &d: devices){
                std::cout << d.getInfo<CL_DEVICE_NAME>() << std::endl;
        }
        //Choose the default compute device
        cl::Device mydevice=devices[0];
        
        //Create a program context and a simple OpenCL source
        cl::Context context({mydevice});
        cl::Program::Sources sources;

        
        /**
         * The kernel takes the two numbers as A and B, the operator as C and the Output as D. C is encoded as an integer between 0 and 4, while D is double for power function.
         */
        std::string simple_kernel=
                "void kernel adder(global const int*A, global const int *B, global const int *C, global double *D){"
                "   switch(C[get_global_id(0)]){"
                "       case 0: D[get_global_id(0)]=A[get_global_id(0)]+B[get_global_id(0)]; break;"
                "       case 1: D[get_global_id(0)]=A[get_global_id(0)]-B[get_global_id(0)]; break;"
                "       case 2: D[get_global_id(0)]=A[get_global_id(0)]*B[get_global_id(0)]; break;"
                "       case 3: D[get_global_id(0)]=A[get_global_id(0)]/B[get_global_id(0)]; break;"
                "       case 4: D[get_global_id(0)]=(pow((double)(A[get_global_id(0)]),(double)(B[get_global_id(0)]))); break;"
                "   }"
                "}";
        sources.push_back({simple_kernel.c_str(), simple_kernel.length()});
        cl::Program program(context,sources);
        //Compile the program
        if(program.build({mydevice}) != CL_SUCCESS){
                std::cout << program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(mydevice) << std::endl;
                exit(1);
        }
        std::cout << "Sucessfully build the kernel" << std::endl;

        //Create input and output buffers
        cl::Buffer buffer_A(context,CL_MEM_READ_WRITE, sizeof(int)*PROBLEM_SIZE);
    cl::Buffer buffer_B(context,CL_MEM_READ_WRITE, sizeof(int)*PROBLEM_SIZE);
    cl::Buffer buffer_C(context,CL_MEM_READ_WRITE, sizeof(int)*PROBLEM_SIZE);
    //Attention to create D as a double buffer
    cl::Buffer buffer_D(context,CL_MEM_READ_WRITE, sizeof(double)*PROBLEM_SIZE);
    
    //Create local arrays for the inputs
    int *adata=new int[PROBLEM_SIZE];
    int *bdata=new int[PROBLEM_SIZE];
    int *cdata=new int[PROBLEM_SIZE];
    
    //Fill the arrays with identic values, make the operator a number between 0 and 4
    for(unsigned long i=0;i<PROBLEM_SIZE;i++){
        adata[i]=i;
        bdata[i]=i;
        cdata[i]=i%5; //Operand
    }
    
    cl::CommandQueue queue(context,mydevice);
    
    //Trigger a copy of the local buffers to the opencl buffers
    queue.enqueueWriteBuffer(buffer_A,CL_TRUE,0,sizeof(int)*PROBLEM_SIZE,adata);
    queue.enqueueWriteBuffer(buffer_B,CL_TRUE,0,sizeof(int)*PROBLEM_SIZE,bdata);
    queue.enqueueWriteBuffer(buffer_C,CL_TRUE,0,sizeof(int)*PROBLEM_SIZE,cdata);
    
    //Create a kernel to call the "adder" function
    cl::Kernel kernel(program,"adder");
    kernel.setArg(0, buffer_A);
    kernel.setArg(1, buffer_B);
    kernel.setArg(2, buffer_C);
    kernel.setArg(3, buffer_D);
    
    //Set the execution range for the kernel
    cl::NDRange global(PROBLEM_SIZE);
    cl::NDRange local(1);
    
    double before=omp_get_wtime();
    //Execute the kernel and measure the ti,e
    queue.enqueueNDRangeKernel(kernel, cl::NullRange, global, local);
    double end=omp_get_wtime();
    std::cout << "GPU version took " << (end-before)*1000000 << "µs" << std::endl;
    
    //Create a local buffer for the output and copy data from the opencl system
    double *ddata=new double[PROBLEM_SIZE];
    queue.enqueueReadBuffer(buffer_D, CL_TRUE, 0, sizeof(double)*PROBLEM_SIZE,ddata);
    
    //Perform CPU computation to validate the result
    double *correctcdata=new double[PROBLEM_SIZE];
    
    //Compare with the CPU version, use std::pow instead of pow
    before=omp_get_wtime();
#pragma omp parallel for
    for(unsigned long i=0;i<PROBLEM_SIZE;i++){
        switch(cdata[i]){
            case 0: correctcdata[i]=adata[i]+bdata[i]; break;
            case 1: correctcdata[i]=adata[i]-bdata[i]; break;
            case 2: correctcdata[i]=adata[i]*bdata[i]; break;
            case 3: correctcdata[i]=adata[i]/bdata[i]; break;
            case 4: correctcdata[i]=std::pow(adata[i],bdata[i]); break;
        }
    }
    end=omp_get_wtime();
    std::cout << "CPU version took " << (end-before)*1000000 << "µs" << std::endl;
    
    for(unsigned long i=0;i<PROBLEM_SIZE;i++){
        if(correctcdata[i]!=ddata[i]){
            std::cout << "The result is wrong at index " << i << ", expected " << correctcdata[i] << " but found " << ddata[i] << std::endl;
            exit(1);
        }
    }
    std::cout << "The GPU computation is consistent with the CPU computation" << std::endl;
}
